/*
 * Copyright 2024 Browsolate Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package com.browsolate;

import com.fasterxml.jackson.databind.ObjectMapper;
import okhttp3.*;
import software.amazon.awssdk.auth.credentials.EnvironmentVariableCredentialsProvider;
import software.amazon.awssdk.regions.Region;
import software.amazon.awssdk.services.secretsmanager.SecretsManagerClient;
import software.amazon.awssdk.services.secretsmanager.model.GetSecretValueRequest;
import software.amazon.awssdk.services.secretsmanager.model.GetSecretValueResponse;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.net.ssl.SSLSocketFactory;
import java.security.cert.X509Certificate;
import javax.crypto.Cipher;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.nio.charset.StandardCharsets;
import java.security.SecureRandom;
import java.util.Base64;
import java.util.Scanner;
import java.util.Map;
import java.util.HashMap;

public class CreateBrowsolateUrl {

    private static final ObjectMapper objectMapper = new ObjectMapper();

    public static void main(String[] args) throws Exception {
        // Parse arguments
        CommandLineArgs options = CommandLineArgs.parse(args);

        // Check required options
        if (options.baseUrl == null || options.navigateUrl == null || (options.key == null && options.secret == null)) {
            System.err.println("Error: Either --key or --secret must be provided.");
            System.exit(1);
        }

        // Construct the JSON payload
        String jsonData = constructJsonPayload(options.navigateUrl);

        // If using API mode
        if (options.api) {
            if (options.password == null) {
                System.err.println("Error: --password is required in API mode.");
                System.exit(1);
            }
            String encryptedData = apiEncrypt(options.baseUrl, options.key, jsonData, options.password, options.ignore_ssl);
            String finalUrl = String.format("%s/viewer.html?&keyId=%s&b8data=%s", options.baseUrl, options.key, encryptedData);
            System.out.println(finalUrl);
        } else {
            // Local encryption mode
            byte[] keyBytes;
            if (options.secret != null) {
                if (options.region == null) {
                    System.err.println("Error: Region is required when using AWS Secrets Manager.");
                    System.exit(1);
                }
                keyBytes = fetchSecretFromAWS(options.secret, options.region);
            } else {
                keyBytes = Base64.getDecoder().decode(options.key);
            }

            String encryptedData = encrypt(keyBytes, jsonData);
            String finalUrl = String.format("%s/viewer.html?&keyId=default&b8data=%s", options.baseUrl, encryptedData);
            System.out.println("Encrypted URL: " + finalUrl);
        }
    }

    public static String apiEncrypt(String baseUrl, String keyId, String jsonData, String password, boolean ignoreSsl) throws Exception {
        OkHttpClient.Builder clientBuilder = new OkHttpClient.Builder();

        // Optionally ignore SSL verification if ignoreSsl is true
        if (ignoreSsl) {
            // Create a trust manager that does not validate certificate chains
            X509TrustManager trustAllCerts = new X509TrustManager() {
                @Override
                public void checkClientTrusted(X509Certificate[] chain, String authType) {}

                @Override
                public void checkServerTrusted(X509Certificate[] chain, String authType) {}

                @Override
                public X509Certificate[] getAcceptedIssuers() {
                    return new X509Certificate[0];
                }
            };

            // Install the all-trusting trust manager
            SSLContext sslContext = SSLContext.getInstance("SSL");
            sslContext.init(null, new TrustManager[]{trustAllCerts}, new java.security.SecureRandom());
            SSLSocketFactory sslSocketFactory = sslContext.getSocketFactory();

            clientBuilder.sslSocketFactory(sslSocketFactory, trustAllCerts);
            clientBuilder.hostnameVerifier((hostname, session) -> true);
        }


        OkHttpClient client = clientBuilder.build();

        // Construct the URL for the API call
        String apiUrl = String.format("%s:5443/encrypt", baseUrl);

        // Basic authentication header
        String auth = Credentials.basic("b8admin", password);

        Map<String, Object> payload = new HashMap<>();
        payload.put("keyName", keyId);
        payload.put("data", jsonData);

        // Create a request body with the JSON payload
        RequestBody body = RequestBody.create(objectMapper.writeValueAsString(payload), MediaType.get("application/json"));

        // Construct the API request
        Request request = new Request.Builder()
                .url(apiUrl)
                .addHeader("Authorization", auth)  // Add the authorization header
                .addHeader("Content-Type", "application/json")  // Specify the content type
                .post(body)  // Send the request as a POST
                .build();

        // Execute the API request and get the response
        try (Response response = client.newCall(request).execute()) {

            if (!response.isSuccessful()) {
                throw new Exception("Failed to encrypt data via API: " + response);
            }

            // Parse the response body to get the encrypted data
            String responseBody = response.body().string();
            return new ObjectMapper().readTree(responseBody).get("encryptedData").asText();
        }
    }

    // Function to encrypt data using AES-256-CBC
    public static String encrypt(byte[] key, String plaintext) throws Exception {
        // Generate a random 16-byte IV
        SecureRandom random = new SecureRandom();
        byte[] iv = new byte[16];
        random.nextBytes(iv);

        // Set up AES encryption
        Cipher cipher = Cipher.getInstance("AES/CBC/PKCS5Padding");
        SecretKeySpec keySpec = new SecretKeySpec(key, "AES");
        IvParameterSpec ivSpec = new IvParameterSpec(iv);
        cipher.init(Cipher.ENCRYPT_MODE, keySpec, ivSpec);

        // Encrypt data
        byte[] encrypted = cipher.doFinal(plaintext.getBytes(StandardCharsets.UTF_8));
        byte[] combined = new byte[iv.length + encrypted.length];
        System.arraycopy(iv, 0, combined, 0, iv.length);
        System.arraycopy(encrypted, 0, combined, iv.length, encrypted.length);

        // Return Base64-encoded encrypted data
        return Base64.getUrlEncoder().withoutPadding().encodeToString(combined);
    }

    // Function to fetch secret from AWS Secrets Manager
    public static byte[] fetchSecretFromAWS(String secretName, String region) {
        SecretsManagerClient secretsManager = SecretsManagerClient.builder()
                .region(Region.of(region))
                .credentialsProvider(EnvironmentVariableCredentialsProvider.create())
                .build();

        GetSecretValueRequest request = GetSecretValueRequest.builder()
                .secretId(secretName)
                .build();

        GetSecretValueResponse response = secretsManager.getSecretValue(request);

        if (response.secretString() != null) {
            String secretData = response.secretString();
            try {
                String key = objectMapper.readTree(secretData).get("default").asText();
                return Base64.getDecoder().decode(key);
            } catch (Exception e) {
                throw new RuntimeException("Failed to retrieve or parse secret from AWS Secrets Manager", e);
            }
        } else {
            throw new RuntimeException("Secret does not contain a SecretString");
        }
    }

    public static String constructJsonPayload(String jsonUrl) {
        Map<String, Object> payload = new HashMap<>();

        payload.put("clipboardRead", true);
        payload.put("clipboardWrite", true);
        payload.put("readOnly", false);
        payload.put("navigation", true);
        payload.put("urlEditable", false);
        payload.put("sessionLifetime", 300);
        payload.put("sessionExtensionAllowed", false);
        payload.put("sessionExtensionTime", 30);
        payload.put("screenWidth", 1024);
        payload.put("screenHeight", 768);
        payload.put("userAgent", "");
        payload.put("customRequestHeaders", "header1:value1,header2:value2");
        payload.put("debugMode", false);
        payload.put("linkId", "");
        payload.put("linkExpiry", -1);
        payload.put("logDirectUrlNavigation", false);
        payload.put("logIndirectUrlNavigation", false);
        payload.put("url", jsonUrl);

        // Convert the Map to a JSON string using Jackson
        try {
            return objectMapper.writeValueAsString(payload);
        } catch (Exception e) {
            throw new RuntimeException("Failed to construct JSON payload", e);
        }
    }

}
